# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

from csv import Error
import os
from pathlib import Path
import subprocess

from rich.console import Group
from rich.live import Live
from rich.panel import Panel
from rich.table import Table

from vaultio.client import Client

from faker import Faker

faker = Faker()

def iter_items(client):
    folder_map = {f["id"]: f["name"] for f in client.list(type="folder")}

    for item in client.list():
        path = Path(folder_map[item["folderId"]]) / item["name"]
        yield path, item

def pass_insert(path, value):
    subprocess.check_output(["pass", "insert", "-m", path], input=value)

def getpath(entry, value_path):
    for k in value_path.split("/"):
        if (isinstance(k, str) and k in entry) or (isinstance(k, int) and k < len(entry)):
            entry = entry[k]
        else:
            return None
    return entry

def backup_value(entry_path, entry, value_path):
    value = getpath(entry, value_path)
    if value is not None:
        pass_insert(entry_path / value_path, value.encode("utf-8"))

def backup_attachments(client, display, item_path, item):

    for attachment in item.get("attachments", ()):
        attachment_path = item_path / "attachments" / attachment["fileName"]
        pass_insert(attachment_path, client.get_attachment(attachment["id"], item["id"]))
        display.add_backup_attachment(attachment["fileName"], attachment["id"])

def counts(entries):
    cts = {}
    for entry in entries:
        if entry["name"] in cts:
            cts[entry["name"]] += 1
        else:
            cts[entry["name"]] = 1
    return cts

def backup_fields(item_path, item):
    if item.get("fields") is None:
        return

    cts = counts(item["fields"])

    for field in item["fields"]:

        field_path = item_path / "fields" / field["name"]

        if cts[field["name"]] != 1 or field["name"] == "fields" or field["value"] is None:
            # print(f"Ignored field {field_path}")
            continue

        pass_insert(field_path, field["value"].encode())

def backup(client, display, item_path, item):

    backup_value(item_path, item, "id")
    backup_value(item_path, item, "login/username")
    backup_value(item_path, item, "login/password")
    backup_value(item_path, item, "notes")
    backup_fields(item_path, item)
    backup_attachments(client, display, item_path, item)

class Display:

    def __init__(self, live) -> None:
        self.live = live
        self.items = []        # List of (name, uuid)
        self.attachments = []  # List of (filename, uuid)

    def add_backup_item(self, name, uuid):
        name = faker.url()
        uuid = faker.uuid4()
        self.items.append((name, uuid))
        self.items = self.items[-5:]

    def add_backup_attachment(self, fn, uuid):
        fn = faker.file_name()
        uuid = faker.password(length=32, special_chars=False, upper_case=False)
        self.attachments.append((fn, uuid))
        self.attachments = self.attachments[-5:]

    def _build_table(self, rows, col1, col2):
        table = Table.grid(padding=(0, 1))
        table.add_column(col1, style="bold")
        table.add_column(col2, style="dim")

        if rows:
            for a, b in rows:
                table.add_row(a, b)
        else:
            table.add_row("[dim]None yet[/]", "")

        return table

    def update(self, total, idx):

        progress = Table.grid()
        progress.add_column()
        progress.add_row(f"{(idx + 1) / total * 100:.1f}%")

        items_panel = Panel(
            self._build_table(self.items, "Item", "ID"),
            title="Backed Up Items",
            border_style="green"
        )

        attachments_panel = Panel(
            self._build_table(self.attachments, "File", "ID"),
            title="Backed Up Attachments",
            border_style="green"
        )

        progress_panel = Panel(progress, title="Backup Progress", border_style="green")

        layout = Group(items_panel, attachments_panel, progress_panel)
        self.live.update(layout)

def main():
    with Client() as client, Live(refresh_per_second=30) as live:
        display = Display(live)
        client.unlock()

        items = list(iter_items(client))
        cts = counts(item for _, item in items)
        total = len(items)

        for idx, (item_path, item) in enumerate(items):

            if cts[item["name"]] != 1:
                continue

            backup(client, display, Path("bitwarden") / item_path, item)
            display.add_backup_item(item["name"], item["id"])
            display.update(total, idx)

if __name__ == "__main__":
    main()



# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

import itertools
import json
import mimetypes
import os
import re
import shutil
import socket
import subprocess
from pathlib import Path
import time
from email.utils import encode_rfc2231
from urllib.parse import urlencode, quote

from vaultio.util import BW_PATH, CACHE_DIR, SOCK_SUPPORT, kill_process_listening_on_socket

class HttpResponseError(Exception):

    def __init__(self, status, reason, headers, content) -> None:
        self.status = status
        self.reason = reason
        self.headers = headers
        self.content = content
        super().__init__(f"Http Error [{status}]: {reason}")

class HttpResponse:

    def __init__(self, status, reason, headers, chunks):
        self.status = status
        self.reason = reason
        self.headers = headers
        self.chunks = chunks

    def bytes(self, check=False):
        if check: self.check()
        return b"".join(
            chunk for chunk in self.chunks
        )

    def content(self, check=False):
        if check: self.check()
        return "".join(
            chunk.decode() for chunk in self.chunks
        )

    def check(self):
        if self.status == 200:
            return
        else:
            raise HttpResponseError(self.status, self.reason, self.headers, self.content())

    def json(self, check=False):
        return json.loads(self.content(check))

def require_bw():
    if BW_PATH is None:
        Exception("BW CLI not found. Try `vaultio build` to resolve dependencies.")

def require_bw_socks():
    if SOCK_SUPPORT is None:
        Exception("BW CLI supporting socket serve not found. Try `vaultio build ` to resolve dependencies.")

def bw_serve(socks=None, host=None, port=None, sock_path=None, fd=None, **kwds):
    if BW_PATH is None:
        Exception("BW CLI supporting socket serve not found. Try `vaultio build` to resolve dependencies.")
    kill_process_listening_on_socket(sock_path)

    args = [str(CACHE_DIR / "bin" / "bw"), "serve", "--hostname"]

    if socks is not None:
        require_bw_socks()
        fd = socks[1].fileno()
        kwds["pass_fds"] = (fd,)
        args +=  [f"fd+connected://{fd}"]
    elif fd is not None:
        require_bw_socks()
        kwds["pass_fds"] = (fd,)
        args +=  [f"fd+listening://{fd}"]
    elif sock_path is not None:
        require_bw_socks()
        args +=  [f"unix://{sock_path}"]
    else:
        assert host is not None
        args += [str(host)]

    if port is not None:
        args += ["--port", str(port)]

    return subprocess.Popen(
        args,
        **kwds,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL
    )

class Serve:

    def __init__(self, socks=None, host=None, port=None, sock_path=None, fd=None, send_sz=4096, recv_sz=4096, serve=True, wait=True):

        if socks is None and host is None and sock_path is None and fd is None:
            if SOCK_SUPPORT:
                socks = socket.socketpair()
                # sock_dir = Path.home() / ".cache" / "vaultio" / "socket"
                # sock_path = os.path.join(sock_dir, "serve.sock")
            else:
                host = "localhost"
                port = int(8087)

        self.socks = socks
        self.host = host
        self.port = port
        self.sock_path = sock_path
        self.fd = fd

        self.send_sz = send_sz
        self.recv_sz = recv_sz
        self.serve = serve
        self.wait = wait

        self.proc = None
        self.start()

    def serve_socket(self):
        if self.sock_path and os.path.exists(self.sock_path):
            os.unlink(self.sock_path)
        self.proc = bw_serve(self.socks, self.host, self.port, self.sock_path, self.fd)
        return self.proc

    def start(self):

        if self.proc is not None:
            return

        if self.serve:
            self.proc = self.serve_socket()
        else:
            self.proc = None

        if self.wait:
            self.wait_socket()

    def end(self):

        if self.sock_path and os.path.exists(self.sock_path):
            os.unlink(self.sock_path)

        if self.proc is not None:
            self.proc.terminate()
            self.proc.wait()
            self.proc = None

    def connect_socket(self):
        assert self.socks is None
        if self.sock_path is not None:
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.connect(self.sock_path)
        elif self.fd is not None:
            sock = socket.socket(fileno=self.fd)
        else:
            assert self.host is not None
            if self.port is not None:
                sock = socket.create_connection((self.host, self.port))
            else:
                sock = socket.create_connection((self.host,))
        return sock

    def wait_socket(self):
        start = time.time()
        timeout = 5
        interval = .2
        while self.sock_path is not None and not os.path.exists(self.sock_path):
            # Socket path doesn't exist
            time.sleep(interval)
        while self.socks is None:
            try:
                with self.connect_socket():
                    # Socket exists and connected
                    return
            except (ConnectionRefusedError, FileNotFoundError) as e:
                print(e)
                # Socket exists but isn't ready yet
                time.sleep(interval)
                if time.time() - start > timeout:
                    raise TimeoutError(f"Could not connect to socket {self.sock_path} within {timeout} seconds")
            time.sleep(interval)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def cleanup(self):
        # Remove the socket file if it exists
        if self.sock_path:
            if os.path.exists(self.sock_path):
                try:
                    os.remove(self.sock_path)
                except OSError as e:
                    # print(f"Error removing socket: {e}")
                    pass

    def parse_header(self, re_header, bfr):
        m = re_header.match(bfr)

        if m:
            status = int(m.group("status").decode())
            reason = m.group("reason").decode()
            headers = dict(line.split(": ") for line in m.group("headers").decode().splitlines())
            chunk = bfr[m.end():]
            return status, reason, headers, chunk
        else:
            return None

    def request_header(self, sock):

        re_header = re.compile("\r\n".join((
            r"HTTP/1.1 (?P<status>\d{3}) (?P<reason>[^\r\n]+)",
            r"(?P<headers>(?:[^\r\n]+:\s*[^\r\n]*\r\n)*)",
            ""
        )).encode("utf-8"))

        bfr = bytearray()

        while True:
            chunk = sock.recv(self.recv_sz)
            if not chunk:
                raise Exception(f"Couldn't parse header:\n{bfr.decode()}")
            bfr += chunk
            header = self.parse_header(re_header, bfr)
            if header:
                return header

    def recv_chunks(self, sock, chunk, headers):

        if "Content-Length" in headers:
            content_length = int(headers["Content-Length"])
        else:
            content_length = None

        yield chunk

        if content_length is not None:
            content_length -= len(chunk)

        while content_length is None or content_length > 0:
            chunk = sock.recv(self.recv_sz)
            if not chunk:
                break
            # print("RECV", chunk)
            yield chunk
            if content_length is not None:
                content_length -= len(chunk)

    def send_chunks(self, sock, chunks):
        bfr = bytearray()
        for chunk in chunks:
            bfr += chunk
            while len(bfr) >= self.recv_sz:
                # print(bfr[:self.recv_sz])
                sock.sendall(bfr[:self.recv_sz])
                # print("SEND", bfr)
                bfr = bfr[self.recv_sz:]
        if len(bfr):
            # print(bfr)
            # print("SEND", bfr)
            sock.sendall(bfr)

    def _request_connected(self, sock, req_chunks, headers=None):
        self.send_chunks(sock, req_chunks)
        header = self.request_header(sock)
        status, reason, headers, chunk = header
        # print(chunk)
        yield status, reason, headers
        for chunk in self.recv_chunks(sock, chunk, headers):
            # print(chunk)
            yield chunk

    def _request(self, endpoint, method, headers=None, body=None, content_type=None, params=None, content_length=None):

        if params is not None:
            endpoint=f"{endpoint}?{urlencode(params)}"

        if headers is None:
            headers = ""
        else:
            headers = "\r\n" + "\r\n".join(f"{k}: {v}" for k, v in headers.items())

        req_chunks = [
            f"{method} {endpoint} HTTP/1.1",
            "Host: localhost"
        ]

        if self.socks is None:
            req_chunks.append("Connection: disconnect")
        else:
            req_chunks.append("Connection: keep-alive")

        if method in ("POST", "PUT"):

            if content_length is not None:
                content_length = str(content_length)
            else:
                content_length = "0"

            req_chunks.append(f"Content-Length: {content_length}")

            if content_type:
                req_chunks.append(f"Content-Type: {content_type}")

        req_chunks = (("\r\n".join(req_chunks) + "\r\n\r\n").encode("utf-8"),)

        if body is not None:
            req_chunks = itertools.chain(req_chunks, body)

        if self.socks is None:
            with self.connect_socket() as sock:
                for chunk in self._request_connected(sock, req_chunks, headers):
                    yield chunk
        else:
            sock = self.socks[0]
            for chunk in self._request_connected(sock, req_chunks, headers):
                yield chunk
            # self.send_chunks(sock, req_chunks)
            # header = self.request_header(sock)
            # status, reason, headers, chunk = header
            # # print(chunk)
            # yield status, reason, headers
            # for chunk in self.recv_chunks(sock, chunk, headers):
            #     # print(chunk)
            #     yield chunk

    def request(self, endpoint, method, headers=None, body=None, content_type=None, params=None, content_length=None):
        chunks = self._request(endpoint, method, headers, body, content_type, params, content_length)
        status, reason, headers = next(chunks)
        return HttpResponse(status, reason, headers, chunks)

    def request_bytes(self, endpoint, method, headers=None, value=None, params=None):
        if value is None:
            body = None
            content_length = None
        else:
            value = json.dumps(value).encode()
            body = (value,)
            content_length = len(value)
        content_type="application/json"
        resp = self.request(endpoint, method, headers, body, content_type, params, content_length)
        return resp.bytes(check=True)

    def request_text(self, endpoint, method, headers=None, value=None, params=None):
        chunks = self.request_bytes(endpoint, method, headers, value, params)
        return chunks.decode()

    def request_json(self, endpoint, method, headers=None, value=None, params=None, text=False):
        text = self.request_text(endpoint, method, headers, value, params)
        return json.loads(text)

    def file_pre_body(self, fpath, boundary):
        filename = encode_rfc2231(os.path.basename(fpath))
        mime_type, _ = mimetypes.guess_type(fpath) or "application/octet-stream"
        mime_type = mime_type or "application/octet-stream"
        field_name = "file"

        return ("\r\n".join((
            f"--{boundary}",
            f'Content-Disposition: form-data; name="{field_name}"; filename*={filename}',
            f"Content-Type: {mime_type}",
        )) + "\r\n\r\n").encode()

    def file_post_body(self, boundary):
        return f"\r\n--{boundary}--\r\n".encode()

    def file_chunks(self, fpath, pre_body, post_body, file_size):

        yield pre_body

        with open(fpath, "rb") as fp:

            while file_size:

                ed = min(file_size, self.send_sz)
                file_size -= ed
                yield fp.read(ed)

        yield post_body

    def request_file(self, endpoint, method, headers=None, fpath=None, params=None):

        boundary="----PyFormBoundary"
        content_type=f"multipart/form-data; boundary={boundary}"

        if fpath is None:
            body = None
            content_length = 0
        else:
            pre_body = self.file_pre_body(fpath, boundary)
            post_body = self.file_post_body(boundary)
            file_size = os.path.getsize(fpath)
            body = self.file_chunks(fpath, pre_body, post_body, file_size)
            content_length = len(pre_body) + file_size + len(post_body)

        resp = self.request(endpoint, method, headers, body, content_type, params, content_length)
        return resp.json(check=True)



# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
import shutil
import psutil

def kill_process_listening_on_socket(socket_path):
    for proc in psutil.process_iter(['pid', 'name']):
        try:
            for conn in proc.net_connections(kind='unix'):
                if conn.laddr == socket_path:
                    print(f"Killing PID {proc.pid} ({proc.name()}) listening on {socket_path}")
                    proc.terminate()
                    proc.wait(timeout=2)
                    return True
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            continue
    print(f"No process found listening on {socket_path}")
    return False

CACHE_DIR = Path.home() / ".cache" / "vaultio"

if (CACHE_DIR / "bin" / "bw").exists():
    SOCK_SUPPORT = True
    BW_PATH = CACHE_DIR / "bin" / "bw"
else:
    SOCK_SUPPORT = False
    BW_PATH = shutil.which("bw")

def remove_none(value):
    ret = {k: v for k, v in value.items() if v is not None}
    return ret or None

def password_input():
    import tkinter as tk
    from tkinter import simpledialog
    root = tk.Tk()
    root.withdraw()
    return simpledialog.askstring("Password", "Enter your password:", show='*')



from csv import Error
import os
from pathlib import Path
import subprocess

from rich.console import Group
from rich.live import Live
from rich.panel import Panel
from rich.table import Table

from pybw.client import Client

def iter_items(client):
    folder_map = {f["id"]: f["name"] for f in client.list(type="folder")}

    for item in client.list():
        path = Path(folder_map[item["folderId"]]) / item["name"]
        yield path, item

def pass_insert(path, value):
    subprocess.check_output(["pass", "insert", "-m", path], input=value)

def getpath(entry, value_path):
    for k in value_path.split("/"):
        if (isinstance(k, str) and k in entry) or (isinstance(k, int) and k < len(entry)):
            entry = entry[k]
        else:
            return None
    return entry

def backup_value(entry_path, entry, value_path):
    value = getpath(entry, value_path)
    if value is not None:
        pass_insert(entry_path / value_path, value.encode())

def backup_attachments(client, display, item_path, item):
    for attachment in item.get("attachments", ()):
        attachment_path = item_path / "attachments" / attachment["fileName"]
        pass_insert(attachment_path, client.get_attachment(attachment["id"], item["id"]))
        display.add_backup_attachment(attachment["fileName"], attachment["id"])

def name_counts(entries):
    '''Count the occurrences of .name'''
    cts = {}
    for entry in entries:
        if entry["name"] in cts:
            cts[entry["name"]] += 1
        else:
            cts[entry["name"]] = 1
    return cts

def backup_fields(item_path, item):

    if item.get("fields") is None:
        return

    cts = name_counts(item["fields"])

    for field in item["fields"]:

        field_path = item_path / "fields" / field["name"]

        if cts[field["name"]] != 1 or field["name"] == "fields" or field["value"] is None:
            # print(f"Ignored field {field_path}")
            continue

        pass_insert(field_path, field["value"].encode())

def backup_item(client, display, item_path, item):
    backup_value(item_path, item, "id")
    backup_value(item_path, item, "login/username")
    backup_value(item_path, item, "login/password")
    backup_value(item_path, item, "notes")
    backup_fields(item_path, item)
    backup_attachments(client, display, item_path, item)

class Display:

    def __init__(self, live) -> None:
        self.live = live
        self.items = []        # List of (name, uuid)
        self.attachments = []  # List of (filename, uuid)

    def add_backup_item(self, name, uuid):
        self.items.append((name, uuid))
        self.items = self.items[-5:]

    def add_backup_attachment(self, fn, uuid):
        self.attachments.append((fn, uuid))
        self.attachments = self.attachments[-5:]

    def _build_table(self, rows, col1, col2):
        table = Table.grid(padding=(0, 1))
        table.add_column(col1, style="bold")
        table.add_column(col2, style="dim")

        if rows:
            for a, b in rows:
                table.add_row(a, b)
        else:
            table.add_row("[dim]None yet[/]", "")

        return table

    def update(self, total, idx):

        progress = Table.grid()
        progress.add_column()
        progress.add_row(f"{(idx + 1) / total * 100:.1f}%")

        items_panel = Panel(
            self._build_table(self.items, "Item", "ID"),
            title="Backed Up Items",
            border_style="green"
        )

        attachments_panel = Panel(
            self._build_table(self.attachments, "File", "ID"),
            title="Backed Up Attachments",
            border_style="green"
        )

        progress_panel = Panel(progress, title="Backup Progress", border_style="green")

        layout = Group(items_panel, attachments_panel, progress_panel)
        self.live.update(layout)

def main():
    with Client() as client, Live(refresh_per_second=30) as live:
        display = Display(live)
        client.unlock()

        items = list(iter_items(client))
        cts = name_counts(item for _, item in items)
        total = len(items)

        for idx, (item_path, item) in enumerate(items):

            if cts[item["name"]] != 1:
                # Skip duplicates
                continue

            backup_item(client, display, Path("bitwarden") / item_path, item)

            # Update display
            display.add_backup_item(item["name"], item["id"])
            display.update(total, idx)

if __name__ == "__main__":
    main()



# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

import platform
import shutil
import subprocess
from pathlib import Path
from importlib.resources import files
from rich.console import Console
from rich.markup import escape

from vaultio.util import CACHE_DIR

console = Console()

def log_step(msg):
    console.log(f":hammer_and_wrench: [bold]{escape(msg)}[/bold]")

def log_info(msg):
    console.log(f":information: [cyan]{escape(msg)}[/cyan]")

def log_done(msg):
    console.log(f":white_check_mark: [green]{escape(msg)}[/green]")

def log_download(msg):
    console.log(f":arrow_down: [bold blue]{escape(msg)}[/bold blue]")

def log_clone(msg):
    console.log(f":inbox_tray: [magenta]{escape(msg)}[/magenta]")

def log_move(msg):
    console.log(f":package: [yellow]{escape(msg)}[/yellow]")

def build_nodeenv():
    log_download("⬇️  Installing Node and npm into the virtualenv...")
    subprocess.run(
        # ["nodeenv", "-p", str(venv), "-n", "20.18.0", "--npm", "none"],
        ["nodeenv", "-p", "-n", "20.18.0", "--npm", "none"],
        check=False,
        text=True
    )

def clone_bw():

    root_dir = CACHE_DIR / "clients"
    cli_dir = root_dir / "apps" / "cli"

    repo_url = "https://github.com/Game4Move78/clients"

    log_clone(f"Cloning Bitwarden CLI from {repo_url}")
    subprocess.run(["git", "clone", repo_url, str(root_dir), "--depth", "1", "--single-branch", "--branch", "feat/unix-socket-support"], check=True)

    assert root_dir.exists()
    assert cli_dir.exists()

    return root_dir, cli_dir

def get_pkg_script():
    system = platform.system()
    machine = platform.machine()

    if system == "Linux":
        return "package:oss:lin", "dist/oss/linux/bw"
    elif system == "Darwin":
        if machine == "arm64":
            return "package:oss:mac-arm64", "dist/oss/macos-arm64/bw"
        else:
            return "package:oss:mac", "dist/oss/macos/bw"
    elif system == "Windows":
        return "package:oss:win", "dist/oss/windows/bw.exe"
    else:
        raise RuntimeError(f"Unsupported platform: {system} ({machine})")

PACKAGE_SCRIPT, BW_PATH = get_pkg_script()

def build_bw():

    root_dir, cli_dir = clone_bw()

    log_done(f"Bitwarden CLI repo found in {root_dir}")

    assert cli_dir.exists(), f"Expected BW directory not found at: {cli_dir}"
    assert cli_dir.exists(), f"Expected BW CLI directory not found at: {cli_dir}"

    log_download("Installing Bitwarden CLI dependencies...")
    subprocess.run(["npm", "install", "-w", "@bitwarden/cli", "--ignore-scripts"], cwd=root_dir, check=True)
    subprocess.run(["npm", "install", "-w", "@bitwarden/cli", "--ignore-scripts"], cwd=cli_dir, check=True)
    subprocess.run(["npm", "install", "-D", "cross-env", "webpack", "tsconfig-paths-webpack-plugin"], cwd=root_dir, check=True)

    log_step("Building Bitwarden CLI...")
    subprocess.run(["npm", "run", "build:oss:prod", "-w", "@bitwarden/cli"], cwd=root_dir, check=True)
    subprocess.run(["npm", "run", PACKAGE_SCRIPT, "-w", "@bitwarden/cli"], cwd=root_dir, check=True)

    if (cli_dir / BW_PATH).exists():
        print(f"{(cli_dir / BW_PATH)} exists. Removing...")
        shutil.rmtree(root_dir)
    log_move(f"Moving {BW_PATH} from {cli_dir} into {CACHE_DIR}")
    shutil.move(cli_dir / BW_PATH, CACHE_DIR / "bin")

    log_info(f"Cleaning up {root_dir}")

    shutil.rmtree(root_dir)

def build(unofficial=True):
    if unofficial:
        log_info("Unofficial build mode: building Bitwarden CLI fork from source.")
        build_nodeenv()
        build_bw()
    else:
        log_info("Official mode: Installing Bitwarden CLI via npm.")
        if not shutil.which("npm"):
            build_nodeenv()
        log_download(f"Installing @bitwarden/cli globally with prefix {CACHE_DIR}")
        subprocess.run(["npm", "install", "@bitwarden/cli", "-g", "--prefix", CACHE_DIR], check=True)
        log_done("Installation complete.")

def has_bw():
    return shutil.which("node") and (CACHE_DIR / "build").exists()



# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

from vaultio.server import Serve

def remove_none(value):
    ret = {k: v for k, v in value.items() if v is not None}
    return ret or None

def password_input():
    import tkinter as tk
    from tkinter import simpledialog
    root = tk.Tk()
    root.withdraw()
    return simpledialog.askstring("Password", "Enter your password:", show='*')

class Client:

    def __init__(self, socks=None, host=None, port=None, sock_path=None, fd=None, serve=True, wait=True, allow_write=True) -> None:
        self._serve = Serve(socks=socks, host=host, port=port, sock_path=sock_path, fd=fd, serve=serve, wait=wait)
        self.allow_write = allow_write

    def __enter__(self):
        self._serve.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._serve.end()

    def serve(self):
        self._serve.serve_socket()
        self._serve.wait_socket()

    def close(self):
        self._serve.end()

    def lock(self):
        value = self._serve.request_json("/lock", "POST")
        return value["success"]

    def unlock(self, password=None):
        if password is None:
            # import getpass
            # password = getpass.getpass("Enter master password: ")
            password = password_input()
        value = self._serve.request_json("/unlock", "POST", value={"password": password})
        return value["data"]["raw"] if value["success"] else None

    def sync(self):
        value = self._serve.request_json("/sync", "POST")
        return value["success"]

    def status(self):
        value = self._serve.request_json("/status", "GET")
        return value["data"]["template"] if value["success"] else None

    def generate(self, length=None, uppercase=None, lowercase=None, numbers=None, special=None, passphrase=None, words=None, seperator=None, capitalize=None, include_number=None):

        params = remove_none(dict(length=length, uppercase=uppercase, lowercase=lowercase, numbers=numbers, special=special, passphrase=passphrase, words=words, seperator=seperator, capitalize=capitalize, include_number=include_number))

        value = self._serve.request_json("/generate", "GET", params=params)

        return value["data"]["data"]

    def fingerprint(self):
        value = self._serve.request_json("/object/fingerprint/me", "GET")
        return value["data"] if value["success"] else None

    def template(self, type):
        value = self._serve.request_json(f"/object/template/{type}", "GET")
        return value["data"]["template"] if value["success"] else None

    def get_attachment(self, attachment_id, item_id):
        params = dict(itemid=item_id)
        value = self._serve.request_bytes(f"/object/attachment/{attachment_id}", "GET", params=params)
        return value

    def new_attachment(self, uuid, fpath=None):
        assert self.allow_write
        params = dict(itemid=uuid)
        value = self._serve.request_file(f"/attachment", "POST", fpath=fpath, params=params)
        return value["data"] if value["success"] else None

    GET_TYPES = {
        "uri",
        "totp",
        "notes",
        "exposed",
        "password",
        "username",
        "item",
        "folder",
        # ...
    }

    def get(self, uuid, type="item"):
        assert type in self.GET_TYPES
        value = self._serve.request_json(f"/object/{type}/{uuid}", "GET")
        return value["data"] if value["success"] else None

    NEW_TYPES = {
        "item",
        "folder",
        # ...
    }

    def new(self, item, type="item"):
        assert self.allow_write
        if type == "send":
            assert item.get("type") == 0
        value = self._serve.request_json(f"/object/{type}", "GET", value=item)
        return value["data"] if value["success"] else None

    def edit(self, value, type="item"):
        assert self.allow_write
        if type == "send":
            assert value.get("type") == 0
        uuid = value["uuid"]
        value = self._serve.request_json(f"/object/{type}/{uuid}", "PUT", value=value)
        return value["data"] if value["success"] else None

    def delete(self, uuid, type="item"):
        assert self.allow_write
        value = self._serve.request_json(f"/object/{type}/{uuid}", "DELETE")
        return value["success"]

    def restore(self, uuid):
        assert self.allow_write
        value = self._serve.request_json(f"/restore/item/{uuid}", "POST")
        return value["success"]

    def list(self, organization_id=None, collection_id=None, folder_id=None, url=None, trash=None, search=None, type="item"):
        if type.rstrip("s") == "item":
            params = remove_none(dict(organizationId=organization_id, collectionId=collection_id, folderId=folder_id, url=url, trash=trash, search=search))
        else:
            params = remove_none(dict(search=search))
        if type.rstrip("s") == "send":
            type = "send"
        else:
            type = type.rstrip("s") + "s"
        value = self._serve.request_json(f"/list/object/{type}", "GET", params=params)
        return value["data"]["data"] if value["success"] else None

    def confirm(self, uuid, organization_id):
        assert self.allow_write
        params = dict(organizationId=organization_id)
        value = self._serve.request_json(f"/confirm/org-member/{uuid}", "POST", params=params)
        return value

    def move(self, item_id, organization_id, collection_ids):
        assert self.allow_write
        value = self._serve.request_json(f"/move/{item_id}/{organization_id}", "POST", value=collection_ids)
        return value

    def pending(self, organization_id):
        assert self.allow_write
        value = self._serve.request_json(f"/device-approval/{organization_id}", "GET")
        #TODO: Check the return structure
        return value

    def trust(self, organization_id, request_id=None):
        assert self.allow_write
        if request_id is None:
            value = self._serve.request_json(f"/device-approval/{organization_id}/approve-all", "POST")
        else:
            value = self._serve.request_json(f"/device-approval/{organization_id}/approve/{request_id}", "POST")
        return value["success"]

    def deny(self, organization_id, request_id=None):
        assert self.allow_write
        if request_id is None:
            value = self._serve.request_json(f"/deny-approval/{organization_id}/deny-all", "GET")
        else:
            value = self._serve.request_json(f"/device-approval/{organization_id}/deny/{request_id}", "POST")
        return value["success"]



# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.


import os
from pathlib import Path
import fire
from vaultio.build import build
from vaultio.client import Client
from vaultio.server import HttpResponse, HttpResponseError
from vaultio.util import SOCK_SUPPORT

class CLI(Client):

    def __init__(self) -> None:
        if SOCK_SUPPORT:
            sock_dir = Path.home() / ".cache" / "vaultio" / "socket"
            sock_path = os.path.join(sock_dir, "serve.sock")
            super().__init__(sock_path=sock_path, serve=False, wait=False)
        else:
            host = "localhost"
            port = int(8087)
            super().__init__(host=host, port=port, serve=False, wait=False)

    def build(self, unofficial=True):
        build(unofficial)

def main():
    try:
        fire.Fire(CLI())
    except HttpResponseError as err:
        print(err.reason, err.content)

if __name__ == '__main__':
    main()



# This file is part of pybw.
#
# pybw is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pybw is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pybw.  If not, see <https://www.gnu.org/licenses/>.

import cmd
from socket import socketpair
from pybw import Client, build
from pybw.server import HttpResponseError

def safe_cmd(func):
    def wrapper(self, arg):
        try:
            return func(self, arg)
        except HttpResponseError as e:
            print(f"[{e.reason}] {e.content}")
    return wrapper

class Shell(cmd.Cmd):
    intro = "Welcome to the pybw shell. Type help or ? to list commands.\n"
    prompt = "(pybw) "

    def __init__(self, **kwargs):
        super().__init__()
        self.client = Client(**kwargs)
        self.client.__enter__()

    def do_build(self, unofficial=False):
        """Build bw cli: build"""
        build(official)

    @safe_cmd
    def do_unlock(self, arg):
        """Unlock the vault: unlock"""
        result = self.client.unlock()
        print("Unlocked" if result else "Failed to unlock")

    @safe_cmd
    def do_lock(self, arg):
        """Lock the vault: lock"""
        success = self.client.lock()
        print("Locked" if success else "Failed to lock")

    @safe_cmd
    def do_list(self, arg):
        """List items: list"""
        items = self.client.list()
        for item in items:
            print(item)

    @safe_cmd
    def do_generate(self, arg):
        """Generate a password: generate"""
        print(self.client.generate())

    @safe_cmd
    def do_status(self, arg):
        """Show vault status: status"""
        print(self.client.status())

    @safe_cmd
    def do_exit(self, arg):
        """Exit the shell: exit"""
        return self._exit()

    @safe_cmd
    def do_quit(self, arg):
        """Exit the shell: quit"""
        return self._exit()

    @safe_cmd
    def _exit(self):
        self.client.__exit__(None, None, None)
        return True

if __name__ == '__main__':
    Shell(socks=socketpair()).cmdloop()
